@@ -20,9 +20,9 @@
 
 #include <QtCore/QDebug>
 
-QT_BEGIN_NAMESPACE
+QT_REQUIRE_CONFIG(draganddrop);
 
-#ifndef QT_NO_DRAGANDDROP
+QT_BEGIN_NAMESPACE
 
 class QWindow;
 class QPlatformWindow;
@@ -33,25 +33,29 @@ class QXcbScreen;
 class QDrag;
 class QShapedPixmapWindow;
 
-class QXcbDrag : public QXcbObject, public QBasicDrag
+class QXcbDrag : public QXcbObject, public QBasicDrag, public QXcbWindowEventListener
 {
 public:
     QXcbDrag(QXcbConnection *c);
     ~QXcbDrag();
 
-    QMimeData *platformDropData() override;
     bool eventFilter(QObject *o, QEvent *e) override;
 
     void startDrag() override;
     void cancel() override;
-    void move(const QPoint &globalPos) override;
-    void drop(const QPoint &globalPos) override;
+    void move(const QPoint &globalPos, Qt::MouseButtons b, Qt::KeyboardModifiers mods) override;
+    void drop(const QPoint &globalPos, Qt::MouseButtons b, Qt::KeyboardModifiers mods) override;
     void endDrag() override;
 
+    Qt::DropAction defaultAction(Qt::DropActions possibleActions, Qt::KeyboardModifiers modifiers) const override;
+
+    void handlePropertyNotifyEvent(const xcb_property_notify_event_t *event) override;
+
     void handleEnter(QPlatformWindow *window, const xcb_client_message_event_t *event, xcb_window_t proxy = 0);
     void handlePosition(QPlatformWindow *w, const xcb_client_message_event_t *event);
     void handleLeave(QPlatformWindow *w, const xcb_client_message_event_t *event);
-    void handleDrop(QPlatformWindow *, const xcb_client_message_event_t *event);
+    void handleDrop(QPlatformWindow *, const xcb_client_message_event_t *event,
+                    Qt::MouseButtons b = { }, Qt::KeyboardModifiers mods = { });
 
     void handleStatus(const xcb_client_message_event_t *event);
     void handleSelectionRequest(const xcb_selection_request_event_t *event);
@@ -66,23 +70,32 @@ public:
 protected:
     void timerEvent(QTimerEvent* e) override;
 
+    bool findXdndAwareTarget(const QPoint &globalPos, xcb_window_t *target_out);
+
 private:
     friend class QXcbDropData;
 
     void init();
 
-    void handle_xdnd_position(QPlatformWindow *w, const xcb_client_message_event_t *event);
+    void handle_xdnd_position(QPlatformWindow *w, const xcb_client_message_event_t *event,
+                              Qt::MouseButtons b = { }, Qt::KeyboardModifiers mods = { });
     void handle_xdnd_status(const xcb_client_message_event_t *event);
     void send_leave();
 
     Qt::DropAction toDropAction(xcb_atom_t atom) const;
+    Qt::DropActions toDropActions(const QVector<xcb_atom_t> &atoms) const;
     xcb_atom_t toXdndAction(Qt::DropAction a) const;
 
+    void readActionList();
+    void setActionList(Qt::DropAction requestedAction, Qt::DropActions supportedActions);
+    void startListeningForActionListChanges();
+    void stopListeningForActionListChanges();
+
     QPointer<QWindow> initiatorWindow;
     QPointer<QWindow> currentWindow;
     QPoint currentPosition;
 
-    QXcbDropData *dropData;
+    QXcbDropData *m_dropData;
     Qt::DropAction accepted_drop_action;
 
     QWindow *desktop_proxy;
@@ -105,6 +118,9 @@ private:
     bool dropped;
     bool canceled;
 
+    // A window from Unity DnD Manager, which does not respect the XDnD spec
+    xcb_window_t xdndCollectionWindow = XCB_NONE;
+
     // top-level window we sent position to last.
     xcb_window_t current_target;
     // window to send events to (always valid if current_target)
@@ -118,6 +134,9 @@ private:
 
     QVector<xcb_atom_t> drag_types;
 
+    QVector<xcb_atom_t> current_actions;
+    QVector<xcb_atom_t> drop_actions;
+
     struct Transaction
     {
         xcb_timestamp_t timestamp;
@@ -139,8 +158,6 @@ private:
 };
 Q_DECLARE_TYPEINFO(QXcbDrag::Transaction, Q_MOVABLE_TYPE);
 
-#endif // QT_NO_DRAGANDDROP
-
 QT_END_NAMESPACE
 
 #endif
